# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Dominik Zuercher

import os
import math
import sys
import shutil
import datetime
import subprocess
import shlex
import portalocker
import multiprocessing
from functools import partial
import numpy as np
import time
from ekit import logger as logger_utils

LOGGER = logger_utils.init_logger(__name__)
TIMEOUT_MESSAGE = (
    "Maximum number of pending jobs reached, "
    "will sleep for 30 minutes and retry"
)


def decimal_hours_to_str(dec_hours):
    """Transforms decimal hours into the hh:mm format

    :param dec_hours: decimal hours, float or int
    :return: string in the format hh:mm
    """

    full_hours = math.floor(dec_hours)
    minutes = math.ceil((dec_hours - full_hours) * 60)

    if minutes == 60:
        full_hours += 1
        minutes = 0

    if minutes < 10:
        time_str = "{}:0{}".format(full_hours, minutes)
    else:
        time_str = "{}:{}".format(full_hours, minutes)

    return time_str


def make_resource_string(
    function,
    main_memory,
    main_time,
    main_scratch,
    main_n_cores,
    watchdog_memory,
    watchdog_time,
    watchdog_scratch,
    watchdog_n_cores,
    merge_memory,
    merge_time,
    merge_scratch,
    merge_n_cores,
    system,
    verbosity=3,
):
    """
    Creates the part of the submission string which handles
    the allocation of ressources

    :param function: The name of the function defined
                     in the executable that will be submitted
    :param main_memory: Memory per core to allocate for the main job
    :param main_time: The Wall time requested for the main job
    :param main_scratch: Scratch per core to allocate for the main job
    :param main_n_cores: Number of cores to allocate for the main job
    :param watchdog_memory: Memory per core to allocate for the watchdog job
    :param watchdog_time: The Wall time requested for the watchdog job
    :param watchdog_scratch: Scratch to allocate for the watchdog job
    :param watchdog_n_cores: Number of cores to allocate for the watchdog job
    :param merge_memory: Memory per core to allocate for the merge job
    :param merge_time: The Wall time requested for the merge job
    :param merge_scratch: Scratch to allocate for the merge jo
    :param merge_n_cores: Number of cores to allocate for the merge job
    :param system: The type of the queing system of the cluster
    :param verbosity: Verbosity level (0 - 4).
    :return: A string that is part of the submission string.
    """

    logger_utils.set_logger_level(LOGGER, verbosity)

    if function == "main":
        mem = main_memory
        time = main_time
        scratch = main_scratch
        n_cores = main_n_cores
    elif function == "watchdog":
        mem = watchdog_memory
        time = watchdog_time
        scratch = watchdog_scratch
        n_cores = watchdog_n_cores
    elif function == "merge":
        mem = merge_memory
        time = merge_time
        scratch = merge_scratch
        n_cores = merge_n_cores
    elif function == "rerun_missing":
        mem = main_memory
        time = main_time
        scratch = main_scratch
        n_cores = main_n_cores
    elif function == "merge_log_files":
        mem = main_memory
        time = 4
        scratch = main_scratch
        n_cores = 1

    if system == "bsub":
        resource_string = (
            "-W {} -R rusage[mem={}] "
            "-R rusage[scratch={}] "
            "-n {}"
            .format(
                decimal_hours_to_str(time),
                mem,
                scratch,
                n_cores
            )
        )
    elif system == "slurm":
        resource_string = (
            "#SBATCH --time={}:00 \n"
            "#SBATCH --mem-per-cpu={} \n"
            "#SBATCH --tmp={} \n"
            "#SBATCH --cpus-per-task={} \n\n".format(
                decimal_hours_to_str(time),
                int(mem),
                int(scratch),
                n_cores
            )
        )
    elif system == "daint":
        resource_string = (
            "#SBATCH --time={}:00 \n#SBATCH --mem={} \n \n".format(
                decimal_hours_to_str(time), mem
            )
        )
        # TODO: local scratch for slurm
        if scratch > 0:
            LOGGER.warning(
                "Not Implemented Warning: Automatic local scratch "
                "allocation not supported for DAINT system. Ignoring."
            )
    LOGGER.debug(f"Built resource string as {resource_string}")

    return resource_string


def get_log_filenames(log_dir, job_name, function, system="bsub"):
    """
    Builds the filenames of the stdout and stderr log files for a
    given job name and a given function to run.

    :param log_dir: directory where the logs are stored
    :param job_name: Name of the job that will write to the log files
    :param function: Function that will be executed
    :param system: The type of the queing system of the cluster
    :return: filenames for stdout and stderr logs
    """
    job_name_ext = job_name + "_" + function
    if system == "slurm" and (function == "main" or function == "main_r"):
        stdout_log = os.path.join(log_dir, "{}_index%a.o".format(job_name_ext))
        stderr_log = os.path.join(log_dir, "{}_index%a.e".format(job_name_ext))
    else:
        stdout_log = os.path.join(log_dir, "{}.o".format(job_name_ext))
        stderr_log = os.path.join(log_dir, "{}.e".format(job_name_ext))
    return stdout_log, stderr_log


def get_source_cmd(source_file, verbosity=3):
    """
    Builds the command to source a given file if the file exists,
    otherwise returns an empty string.

    :param source_file: path to the (possibly non-existing) source file,
                        can be relative and can contain "~"
    :param verbosity: Verbosity level (0 - 4).
    :return: command to source the file if it exists or empty string
    """

    logger_utils.set_logger_level(LOGGER, verbosity)

    source_file_abs = os.path.abspath(os.path.expanduser(source_file))

    if os.path.isfile(source_file_abs):
        source_cmd = "source {}; ".format(source_file_abs)
        LOGGER.debug(f"Running source script at {source_file_abs}")
    else:
        LOGGER.warning(
            "Source file {} not found, skipping".format(source_file)
        )
        source_cmd = ""

    return source_cmd


def get_dependency_string(
    function, jobids, ext_dependencies, system, verbosity=3
):
    """
    Constructs the dependency string which handles which other jobs
    this job is dependent on.

    :param function: The type o function to submit
    :param jobids: Dictionary of the jobids for each job already submitted
    :param ext_dependencies: If external dependencies are given they get
                             added to the dependency string
                             (this happens if epipe is used)
    :param system: The type of the queing system of the cluster
    :param verbosity: Verbosity level (0 - 4).
    :return: A sting which is used as a substring for the submission string
             and it handles the dependencies of the job
    """
    logger_utils.set_logger_level(LOGGER, verbosity)

    dep_string = ""
    # no dependencies for main
    if function == "main":
        if ext_dependencies != "":
            dep_string = '-w "' + ext_dependencies + '"'
        else:
            dep_string = ""
            return dep_string
        if system == "slurm":
            dep_string = dep_string.replace("ended(", "afterany:")
            dep_string = dep_string.replace("started(", "after:")
            dep_string = dep_string.replace(") && ", ",")
            dep_string = dep_string.replace(")", "")
            dep_string = dep_string.replace('-w "', '--dependency="')
        return dep_string
    # watchdog starts along with main
    elif function == "watchdog":
        if "main" in jobids.keys():
            for id in jobids["main"]:
                dep_string += "{}({}) && ".format("started", id)
        else:
            LOGGER.warning(
                "Function {} has not been submitted -> Skipping "
                "in dependencies for {}".format("main", function)
            )
    # rerun missing starts after main
    elif function == "rerun_missing":
        if "main" in jobids.keys():
            for id in jobids["main"]:
                dep_string += "{}({}) && ".format("ended", id)
        else:
            LOGGER.warning(
                "Function {} has not been submitted -> Skipping "
                "in dependencies for {}".format("main", function)
            )
    # merge_log_files starts after main or rerun_missing
    elif function == "merge_log_files":
        if "rerun_missing" in jobids.keys():
            for id in jobids["rerun_missing"]:
                dep_string += "{}({}) && ".format("ended", id)
        elif "main" in jobids.keys():
            for id in jobids["main"]:
                dep_string += "{}({}) && ".format("ended", id)
    # merge starts after all the others
    elif function == "merge":
        if "main" in jobids.keys():
            for id in jobids["main"]:
                dep_string += "{}({}) && ".format("ended", id)
        else:
            LOGGER.warning(
                "Function {} has not been submitted -> Skipping "
                "in dependencies for {}".format("main", function)
            )
        if "watchdog" in jobids.keys():
            for id in jobids["watchdog"]:
                dep_string += "{}({}) && ".format("ended", id)
        else:
            LOGGER.warning(
                "Function {} has not been submitted -> Skipping "
                "in dependencies for {}".format("watchdog", function)
            )
        if "rerun_missing" in jobids.keys():
            for id in jobids["rerun_missing"]:
                dep_string += "{}({}) && ".format("ended", id)
        else:
            LOGGER.warning(
                "Function {} has not been submitted -> Skipping "
                "in dependencies for {}".format("rerun_missing", function)
            )
    else:
        raise ValueError(
            "Dependencies for function" " {} not defined".format(function)
        )
    # remove trailing &&
    if len(dep_string) > 0:
        dep_string = dep_string[:-4]
    if ext_dependencies != "":
        dep_string = dep_string + " && " + ext_dependencies
        # remove leading &&
        if dep_string[:4] == " && ":
            dep_string = dep_string[4:]
    dep_string = '-w "' + dep_string + '"'

    if system == "slurm":
        dep_string = dep_string.replace("ended(", "afterany:")
        dep_string = dep_string.replace("started(", "after:")
        dep_string = dep_string.replace(") && ", ",")
        dep_string = dep_string.replace(")", "")
        dep_string = dep_string.replace('-w "', '--dependency="')

    if len(dep_string) > 0:
        LOGGER.debug(f"Built dependency string as {dep_string}")
    return dep_string


def make_cmd_string(
    function,
    source_file,
    n_jobs,
    tasks,
    mode,
    job_name,
    function_args,
    exe,
    main_memory,
    main_time,
    main_scratch,
    main_n_cores,
    watchdog_time,
    watchdog_memory,
    watchdog_scratch,
    watchdog_n_cores,
    merge_memory,
    merge_time,
    merge_scratch,
    merge_n_cores,
    log_dir,
    dependency,
    system,
    main_name="main",
    batchsize=100000,
    max_njobs=-1,
    add_args="",
    add_bsub="",
    discard_output=False,
    verbosity=3,
    main_mode="jobarray",
    # main_n_cores=1,
    nodes=1,
    MPI_tasks_per_core=1,
    MPI_tasks_per_node=1,
    OpenMP_threads_per_task=1,
):
    """
    Creates the submission string which gets submitted to the queing system

    :param function: The name of the function defined in the
                     executable that will be submitted
    :param source_file: A file which gets executed
                        before running the actual function(s)
    :param n_jobs: The number of jobs that will be requested for the job
    :param tasks: The task string, which will get parsed into the job indices
    :param mode: The mode in which the job will be
                 ran (MPI-job or as a jobarray)
    :param job_name: The name of the job
    :param function_args: The remaining arguments that
                          will be forwarded to the executable
    :param exe: The path of the executable
    :param main_memory: Memory per core to allocate for the main job
    :param main_time: The Wall time requested for the main job
    :param main_scratch: Scratch per core to allocate for the main job
    :param main_n_cores: Number of cores to allocate for the main job
    :param watchdog_memory: Memory per core to allocate for the watchdog job
    :param watchdog_time: The Wall time requested for the watchdog job
    :param watchdog_scratch: Scratch to allocate for the watchdog job
    :param watchdog_n_cores: Number of cores to allocate for the watchdog job
    :param merge_memory: Memory per core to allocate for the merge job
    :param merge_time: The Wall time requested for the merge job
    :param merge_n_cores: Number of cores to allocate for the merge job
    :param log_dir: log_dir: The path to the log directory
    :param merge_scratch: Scratch to allocate for the merge job
    :param dependency: The dependency string
    :param system: The type of the queing system of the cluster
    :param main_name: name of the main function
    :param batchsize: If not zero the jobarray gets divided into batches.
    :param max_njobs: Maximum number of jobs allowed to run at the same time.
    :param add_args: Additional cluster-specific arguments
    :param add_bsub: Additional bsub arguments to pass
    :param discard_output: If True writes stdout/stderr to /dev/null
    :param verbosity: Verbosity level (0 - 4).
    :return: The submission string that wil get submitted to the cluster
    """

    logger_utils.set_logger_level(LOGGER, verbosity)
    LOGGER.info("")

    if function == "main_rerun":
        function = "main"
        if system == "slurm":
            # reruns should use different log files to avoid overwriting
            log_function = "main_r"
        else:
            log_function = "main"
    else:
        log_function = function

    # allocate computing resources
    resource_string = make_resource_string(
        function,
        main_memory,
        main_time,
        main_scratch,
        main_n_cores,
        watchdog_memory,
        watchdog_time,
        watchdog_scratch,
        watchdog_n_cores,
        merge_memory,
        merge_time,
        merge_scratch,
        merge_n_cores,
        system,
        verbosity,
    )

    # get the job name for the submission system and the log files
    job_name_ext = job_name + "_" + function
    stdout_log, stderr_log = get_log_filenames(
        log_dir, job_name, log_function, system
    )

    # construct the string of arguments passed to the executable
    args_string = ""
    for arg in function_args:
        args_string += arg + " "

    # make submission string
    source_cmd = get_source_cmd(source_file, verbosity)

    if mode == "mpi":
        run_cmd = "mpirun python"
    elif mode == "jobarray":
        run_cmd = "python"
    else:
        raise ValueError(f"Run mode {mode} is not known")

    extra_args_string = (
        "--source_file={} --main_memory={} --main_time={} "
        "--main_scratch={} --function={} "
        "--executable={} --n_jobs={} "
        "--log_dir={} --system={} "
        "--main_name={} --batchsize={} --max_njobs={} "
        "--main_n_cores={} "
        '--esub_verbosity={} --main_mode={} --mode={} {}"'.format(
            source_file,
            main_memory,
            main_time,
            main_scratch,
            function,
            exe,
            n_jobs,
            log_dir,
            system,
            main_name,
            batchsize,
            max_njobs,
            main_n_cores,
            verbosity,
            main_mode,
            mode,
            args_string,
        )
    )

    if (function == "main") & (max_njobs > 0):
        max_string = "%{}".format(max_njobs)
    else:
        max_string = ""

    if system == "bsub":

        if n_jobs <= batchsize:
            cmd_string = (
                "bsub -o {} -e {} -J {}[1-{}]{} "
                "{} {} {}"
                ' "{} {} -m esub.submit --job_name={} '
                "--tasks='{}' {}".format(
                    stdout_log,
                    stderr_log,
                    job_name_ext,
                    n_jobs,
                    max_string,
                    resource_string,
                    add_bsub,
                    dependency,
                    source_cmd,
                    run_cmd,
                    job_name,
                    tasks,
                    extra_args_string,
                )
            )
        else:
            LOGGER.warning(
                "You have requested a jobarray with more "
                f"than {batchsize} cores"
                ". Euler cannot handle this. I break down this job into "
                "multiple subarrays and submit them one by one. "
                "Note that this feature currently breakes the rerun "
                "missing capability. Also note that"
                " this process needs to keep running...."
            )

            n_batches = math.ceil(n_jobs / batchsize)
            cmd_string = []
            for rank in range(n_batches):
                if rank < (n_batches - 1):
                    jobs = batchsize
                else:
                    jobs = n_jobs % batchsize
                    if jobs == 0:
                        jobs = batchsize
                first_task = get_indices_splitted(
                    tasks, n_jobs, rank * batchsize
                )
                first_task = first_task[0]
                last_task = get_indices_splitted(
                    tasks, n_jobs, rank * batchsize + jobs - 1
                )
                last_task = last_task[-1]
                tasks_ = f"{first_task} > {last_task + 1}"

                jobname_ = f"{job_name}_{rank}"
                stdout_log_ = stdout_log[:-2] + f"_{rank}.o"
                stderr_log_ = stdout_log[:-2] + f"_{rank}.e"
                cs = (
                    'bsub -o {} -e {} -J {}[1-{}]{} '
                    '{} {} "{} '
                    "{} -m esub.submit --job_name={} --tasks='{}' {}".format(
                        stdout_log_,
                        stderr_log_,
                        job_name_ext,
                        jobs,
                        max_string,
                        resource_string,
                        dependency,
                        source_cmd,
                        run_cmd,
                        jobname_,
                        tasks_,
                        extra_args_string,
                    )
                )
                cmd_string.append(cs)
        if discard_output:
            if isinstance(cmd_string, list):
                for i in range(len(cmd_string)):
                    cmd_string[i] = (
                        cmd_string[i] + " --discard_output &> /dev/null"
                    )
            else:
                cmd_string += " --discard_output &> /dev/null"
    elif system == "slurm":
        # split add_args
        if len(add_args) > 0:
            add_args = add_args.split(",")
        else:
            add_args = []

        LOGGER.warning("SLURM support is in BETA!")

        cmd_string = "sbatch {} submit_{}.slurm".format(
            dependency, job_name_ext
        )
        # write submission file
        with open(f"submit_{job_name_ext}.slurm", "w+") as f:
            f.write("#! /bin/bash \n#\n")
            if discard_output:
                f.write("#SBATCH --output=/dev/null \n")
                f.write("#SBATCH --error=/dev/null \n")
            else:
                f.write("#SBATCH --output={} \n".format(stdout_log))
                f.write("#SBATCH --error={} \n".format(stderr_log))
            f.write("#SBATCH --job-name={} \n".format(job_name_ext))
            for arg in add_args:
                f.write("#SBATCH {} \n".format(arg))
            f.write("#SBATCH --array=1-{}{} \n".format(n_jobs, max_string))
            f.write(resource_string)
            f.write(
                "srun bash; {} {} -m esub.submit --job_name={} "
                "--tasks='{}' {}".format(
                    source_cmd,
                    run_cmd,
                    job_name,
                    tasks,
                    extra_args_string[:-1],
                )
            )
    elif system == "daint":
        # split add_args
        if len(add_args) > 0:
            add_args = add_args.split(",")
        else:
            add_args = []

        cmd_string = "sbatch {} submit_{}.slurm".format(
            dependency, job_name_ext
        )

        # write submission file
        with open(f"submit_{job_name_ext}.slurm", "w+") as f:
            f.write("#! /bin/bash \n#\n")
            if discard_output:
                f.write("#SBATCH --output=/dev/null \n")
                f.write("#SBATCH --error=/dev/null \n")
            else:
                f.write("#SBATCH --output={} \n".format(stdout_log))
                f.write("#SBATCH --error={} \n".format(stderr_log))
            f.write("#SBATCH --job-name={} \n".format(job_name_ext))
            f.write("#SBATCH --constraint=gpu \n")
            f.write("#SBATCH --nodes={} \n".format(nodes))
            f.write(
                "#SBATCH --ntasks-per-core={} \n".format(MPI_tasks_per_core)
            )
            f.write(
                "#SBATCH --ntasks-per-node={} \n".format(MPI_tasks_per_node)
            )
            f.write(
                "#SBATCH --cpus-per-task={} \n".format(OpenMP_threads_per_task)
            )
            for arg in add_args:
                f.write("#SBATCH {} \n".format(arg))
            f.write(resource_string)
            if len(source_cmd) > 0:
                f.write("srun {} \n".format(source_cmd))
            f.write(
                "srun python -m esub.submit --job_name={} "
                "--tasks='{}' {}".format(
                    job_name, tasks, extra_args_string[:-1]
                )
            )

    LOGGER.debug(f"Built total command string as {cmd_string}")
    return cmd_string


def submit_job(
    tasks,
    mode,
    exe,
    log_dir,
    function_args,
    function="main",
    source_file="",
    n_jobs=1,
    job_name="job",
    main_memory=100,
    main_time=1,
    main_scratch=1000,
    main_n_cores=1,
    watchdog_memory=100,
    watchdog_time=1,
    watchdog_scratch=1000,
    watchdog_n_cores=1,
    merge_memory=100,
    merge_time=1,
    merge_scratch=1000,
    merge_n_cores=1,
    dependency="",
    system="bsub",
    main_name="main",
    test=False,
    batchsize=100000,
    max_njobs=100000,
    add_args="",
    add_bsub="",
    discard_output=False,
    verbosity=3,
    main_mode="jobarray",
    keep_submit_files=False,
    nodes=1,
    MPI_tasks_per_core=1,
    MPI_tasks_per_node=1,
    OpenMP_threads_per_task=1,
):
    """
    Based on arguments gets the submission string and submits it to the cluster

    :param tasks: The task string, which will get parsed into the job indices
    :param mode: The mode in which the job will be ran
                 (MPI-job or as a jobarray)
    :param exe: The path of the executable
    :param log_dir: The path to the log directory
    :param function_args: The remaining arguments that will
                          be forwarded to the executable
    :param function: The name of the function defined in the
                     executable that will be submitted
    :param source_file: A file which gets executed before
                        running the actual function(s)
    :param n_jobs: The number of jobs that will be requested for the job
    :param job_name: The name of the job
    :param main_memory: Memory per core to allocate for the main job
    :param main_time: The Wall time requested for the main job
    :param main_scratch: Scratch per core to allocate for the main job
    :param main_n_cores: Number of cores to allocate for the main job
    :param watchdog_memory: Memory per core to allocate for the watchdog job
    :param watchdog_time: The Wall time requested for the watchdog job
    :param watchdog_scratch: Scratch to allocate for the watchdog job
    :param watchdog_n_cores: Number of cores to allocate for the watchdog job
    :param merge_memory: Memory per core to allocate for the merge job
    :param merge_time: The Wall time requested for the merge job
    :param merge_scratch: Scratch to allocate for the merge job
    :param merge_n_cores: Number of cores to allocate for the merge job
    :param dependency: The jobids of the jobs on which this job depends on
    :param system: The type of the queing system of the cluster
    :param main_name: name of the main function
    :param test: If True no submission but just printing submission string to
                 log
    :param batchsize: If number of cores requested is > batchsize, break up
                      jobarrays into jobarrys of size batchsize
    :param max_njobs: Maximum number of jobs allowed to run at the same time
    :param add_args: Additional cluster-specific arguments
    :param add_bsub: Additional bsub arguments to pass
    :param discard_output: If True writes stdout/stderr to /dev/null
    :param verbosity: Verbosity level (0 - 4).
    :param keep_submit_files: If True store SLURM submission files
    :return: The jobid of the submitted job
    """

    logger_utils.set_logger_level(LOGGER, verbosity)

    # assess if number of tasks is valid
    n_tasks = len(get_indices_splitted(tasks, 1, 0))
    if (n_jobs > n_tasks) & (("mpi" not in mode) & (mode != "tasks")):
        raise Exception(
            "You tried to request more jobs than you have tasks. "
            "I assume this is a mistake. Aborting..."
        )
    # get submission string
    cmd_string = make_cmd_string(
        function,
        source_file,
        n_jobs,
        tasks,
        mode,
        job_name,
        function_args,
        exe,
        main_memory,
        main_time,
        main_scratch,
        main_n_cores,
        watchdog_time,
        watchdog_memory,
        watchdog_scratch,
        watchdog_n_cores,
        merge_memory,
        merge_time,
        merge_scratch,
        merge_n_cores,
        log_dir,
        dependency,
        system,
        main_name,
        batchsize,
        max_njobs,
        add_bsub=add_bsub,
        add_args=add_args,
        discard_output=discard_output,
        verbosity=verbosity,
        main_mode=main_mode,
        # main_n_cores=main_n_cores,
        nodes=nodes,
        MPI_tasks_per_core=MPI_tasks_per_core,
        MPI_tasks_per_node=MPI_tasks_per_node,
        OpenMP_threads_per_task=OpenMP_threads_per_task,
    )

    LOGGER.debug(cmd_string)
    if test:
        path_log = get_path_log(log_dir, job_name)
        write_to_log(path_log, cmd_string)
        return []

    # message the system sends if the
    # maximum number of pendings jobs is reached
    msg_limit_reached = "Pending job threshold reached."
    pipe_limit_reached = "stderr"

    if isinstance(cmd_string, str):
        cmd_string = [cmd_string]

    jobids = []
    for cs in cmd_string:
        LOGGER.info("Submitting command:")
        LOGGER.info(cs)
        # submit
        while True:

            output = dict(stdout=[], stderr=[])

            with subprocess.Popen(
                shlex.split(cs),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                bufsize=1,
                universal_newlines=True,
            ) as proc:

                # check for limit concerning maximum number of pending jobs
                if system == "bsub":
                    for line in getattr(proc, pipe_limit_reached):

                        pending_limit_reached = msg_limit_reached in line
                        if pending_limit_reached:
                            break
                        else:
                            output[pipe_limit_reached].append(line)

                    # if the limit has been reached, kill process and sleep
                    if pending_limit_reached:
                        proc.kill()
                        LOGGER.warning(TIMEOUT_MESSAGE)
                        time.sleep(60 * 30)
                        continue

                # read rest of the output
                for line in proc.stdout:
                    output["stdout"].append(line)
                for line in proc.stderr:
                    output["stderr"].append(line)
                break

        # check if process terminated successfully
        if proc.returncode != 0:
            raise RuntimeError(
                'Running the command "{}" failed with'
                "exit code {}. Error: \n{}".format(
                    cmd_string, proc.returncode, "\n".join(output["stderr"])
                )
            )

        # get id of submitted job
        if system == "bsub":
            jobid = output["stdout"][-1].split("<")[1]
            jobid = jobid.split(">")[0]
        elif system == "slurm":
            jobid = output["stdout"][-1].split("job ")[-1]
            if not keep_submit_files:
                for cs in cmd_string:
                    os.remove(f"{cs.split(' ')[-1]}")
        elif system == "daint":
            jobid = output["stdout"][-1].split("job ")[-1]
            if not keep_submit_files:
                for cs in cmd_string:
                    os.remove(f"{cs.split(' ')[-1]}")
        jobids.append(int(jobid))

    LOGGER.info("Submitted job and got jobid(s): {}".format(jobid))
    return jobids


def robust_remove(path):
    """
    Remove a file or directory if existing

    :param path: path to possible non-existing file or directory
    """
    if os.path.isfile(path):
        os.remove(path)
    elif os.path.isdir(path):
        shutil.rmtree(path)
    # recreate
    open(path, "a").close()


def get_path_log(log_dir, job_name):
    """
    Construct the path of the esub log file

    :param log_dir: directory where log files are stored
    :param job_name: name of the job that will be logged
    :return: path of the log file
    """
    path_log = os.path.join(log_dir, job_name + ".log")

    return path_log


def get_path_finished_indices(log_dir, job_name):
    """
    Construct the path of the file containing the finished indices

    :param log_dir: directory where log files are stored
    :param job_name: name of the job for which the indices will be store
    :return: path of the file for the finished indices
    """
    path_finished = os.path.join(log_dir, job_name + "_done.dat")
    return path_finished


def import_executable(exe):
    """
    Imports the functions defined in the executable file.

    :param exe: path of the executable
    :return: executable imported as python module
    """
    sys.path.insert(0, os.path.dirname(exe))
    to_import = os.path.basename(exe).replace(".py", "")
    try:
        executable = __import__(to_import)
    except ImportError:
        raise ImportError(f"Failed to import your executable {exe}")
    return executable


def save_write(path, str_to_write, mode="a"):
    """
    Write a string to a file, with the file being locked in the meantime.

    :param path: path of file
    :param str_to_write: string to be written
    :param mode: mode in which file is opened
    """
    with portalocker.Lock(path, mode=mode, timeout=math.inf) as f:
        # write
        f.write(str_to_write)
        # flush and sync to filesystem
        f.flush()
        os.fsync(f.fileno())


def write_index(index, finished_file):
    """
    Writes the index number on a new line of the
    file containing the finished indices

    :param index: A job index
    :param finished_file: The file to which the
                          jobs will write that they are done
    """
    save_write(finished_file, "{}\n".format(index))


def check_indices(
    indices,
    finished_file,
    exe,
    function_args,
    check_indices_file=True,
    verbosity=3,
):
    """
    Checks which of the indices are missing in
    the file containing the finished indices

    :param indices: Job indices that should be checked
    :param finished_file: The file from which the jobs will be read
    :param exe: Path to executable
    :param check_indices_file: If True adds indices from index file otherwise
                               only use check_missing function
    :param verbosity: Verbosity level (0 - 4).
    :return: Returns the indices that are missing
    """
    if check_indices_file:
        LOGGER.debug("Checking missing file for missing indices...")
        # wait for the indices file to be written
        if os.path.exists(finished_file):
            # first get the indices missing in the log file (crashed jobs)
            done = []
            with open(finished_file, "r") as f:
                for line in f:
                    # Ignore empty lines
                    if line != "\n":
                        done.append(int(line.replace("\n", "")))
            failed = list(set(indices) - set(done))
            LOGGER.debug(f"Found failed indices: {failed}")
        else:
            LOGGER.warning(
                "Did not find File {} -> None of the main functions "
                "recorded its indices. "
                "Not rerunning any jobs".format(finished_file)
            )
            failed = []
    else:
        failed = []

    # if provided use check_missing function
    # (finished jobs but created corrupted output)
    if hasattr(exe, "check_missing"):
        LOGGER.info("Found check_missing function in executable. Running...")
        corrupted = getattr(exe, "check_missing")(indices, function_args)
        LOGGER.debug(f"Found corruped indices: {corrupted}")
    else:
        corrupted = []

    missing = failed + corrupted
    missing = np.unique(np.asarray(missing))
    LOGGER.debug(f"Found failed/corrputed indices: {missing}")
    return missing


def write_to_log(path, line, mode="a"):
    """
    Write a line to a esub log file

    :param path: path of the log file
    :param line: line (string) to write
    :param mode: mode in which the log file will be opened
    """
    extended_line = "{}    {}\n".format(datetime.datetime.now(), line)
    save_write(path, extended_line, mode=mode)


def cd_local_scratch(verbosity=3):
    """
    Change to current working directory to the local scratch if set.

    :param verbosity: Verbosity level (0 - 4).
    """
    if "ESUB_LOCAL_SCRATCH" in os.environ:

        if os.path.isdir(os.environ["ESUB_LOCAL_SCRATCH"]):
            submit_dir = os.getcwd()
            os.chdir(os.environ["ESUB_LOCAL_SCRATCH"])
            os.environ["SUBMIT_DIR"] = submit_dir

            LOGGER.warning(
                "Changed current working directory to {} and "
                "set $SUBMIT_DIR to {}".format(
                    os.getcwd(), os.environ["SUBMIT_DIR"]
                )
            )
        else:
            LOGGER.error(
                "$ESUB_LOCAL_SCRATCH is set to non-existing "
                "directory {}, skipping...".format(
                    os.environ["ESUB_LOCAL_SCRATCH"]
                )
            )
    else:
        LOGGER.debug(
            "Environment variable ESUB_LOCAL_SCRATCH not set. "
            "Not chaning working directory."
        )


def run_local_mpi_job(
    exe, n_cores, function_args, logger, main_name="main", verbosity=3
):
    """
    This function runs an MPI job locally

    :param exe: Path to executable
    :param n_cores: Number of cores
    :param function_args: A list of arguments to be passed to the executable
    :param index: Index number to run
    :param logger: logger instance for logging
    :param main_name: Name of main function in executable
    :param verbosity: Verbosity level (0 - 4).
    :param main_name:
    """
    # construct the string of arguments passed to the executable
    args_string = ""
    for arg in function_args:
        args_string += arg + " "

    # make command string
    cmd_string = (
        "mpirun -np {} python -m esub.submit"
        " --executable={} --tasks='0' --main_name={} "
        "--esub_verbosity={} {}".format(
            n_cores, exe, main_name, verbosity, args_string
        )
    )
    for line in execute_local_mpi_job(cmd_string):
        line = line.strip()
        if len(line) > 0:
            logger.info(line)


def get_indices(tasks):
    """
    Parses the jobids from the tasks string.

    :param tasks: The task string, which will get parsed into the job indices
    :return: A list of the jobids that should be executed
    """
    # parsing a list of indices from the tasks argument
    if ">" in tasks:
        tasks = tasks.split(">")
        start = tasks[0].replace(" ", "")
        stop = tasks[1].replace(" ", "")
        indices = list(range(int(start), int(stop)))
    elif "," in tasks:
        indices = tasks.split(",")
        indices = list(map(int, indices))
    elif os.path.exists(tasks):
        with open(tasks, "r") as f:
            content = f.readline()
        indices = get_indices(content)
    else:
        try:
            indices = [int(tasks)]
        except ValueError:
            raise ValueError("Tasks argument is not in the correct format!")
    return indices


def get_indices_splitted(tasks, n_jobs, rank):
    """
    Parses the jobids from the tasks string.
    Performs load-balance splitting of the jobs and returns the indices
    corresponding to rank. This is only used for job array submission.

    :param tasks: The task string, which will get parsed into the job indices
    :param n_jobs: The number of cores that will be requested for the job
    :param rank: The rank of the core
    :return: A list of the jobids that should
             be executed by the core with number rank
    """

    # Parse
    indices = get_indices(tasks)

    # Load-balanced splitter
    steps = len(indices)
    size = n_jobs
    chunky = int(steps / size)
    rest = steps - chunky * size
    mini = chunky * rank
    maxi = chunky * (rank + 1)
    if rank >= (size - 1) - rest:
        maxi += 2 + rank - size + rest
        mini += rank - size + 1 + rest
    mini = int(mini)
    maxi = int(maxi)

    return indices[mini:maxi]


def function_wrapper(indices, args, func):
    """
    Wrapper that converts a generator to a function.

    :param generator: A generator
    """
    inds = []
    for ii in func(indices, args):
        inds.append(ii)
    return inds


def run_local_tasks(exe, n_jobs, function_args, tasks, function):
    """
    Executes an MPI job locally, running each splitted index list on one core.

    :param exe: The executable from where the main function is imported.
    :param n_jobs: The number of cores to allocate.
    :param function_args: The arguments that
                          will get passed to the main function.
    :param tasks: The indices to run on.
    :param function: The function name to run
    """

    LOGGER.warning(
        "NotImplementedWarning: Using run-tasks creates a multiprocessing "
        "worker pool with just one thread per job. "
        "The n_core arguments are ignored."
    )
    # get executable
    func = getattr(exe, function)

    # Fix function arguments for all walkers
    run_func = partial(function_wrapper, args=function_args, func=func)

    # get splitted indices
    nums = []
    for rank in range(n_jobs):
        nums.append(get_indices_splitted(tasks, n_jobs, rank))

    # Setup mutltiprocessing pool
    pool = multiprocessing.Pool(processes=n_jobs)
    if int(multiprocessing.cpu_count()) < n_jobs:
        raise Exception(
            "Number of CPUs available is smaller \
             than requested number of CPUs"
        )

    # run and retrive the finished indices
    out = pool.map(run_func, nums)
    out = [item for sublist in out for item in sublist]
    return out


def execute_local_mpi_job(cmd_string):
    """
    Execution of local MPI job

    :param cmd_string: The command string to run
    """
    popen = subprocess.Popen(
        shlex.split(cmd_string),
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, cmd_string)



__author__ = 'Dominik Zuercher, Joerg Herbel, Silvan Fischbacher'
__email__ = 'dominik.zuercher@phys.ethz.ch, joerg.herbel@phys.ethz.ch, silvanf@phys.ethz.ch'  # noqa
__credits__ = 'ETH Zurich, Institute for Particle Physics and Astrophysics'

__version__ = "1.10.1"



#! /usr/bin/env python

# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Dominik Zuercher

# System imports
from __future__ import (
    print_function,
    division,
    absolute_import,
    unicode_literals,
)

# package imports
import os
import sys
import argparse
import time
from esub import utils
import subprocess
import shlex
from ekit import logger as logger_utils

LOGGER = logger_utils.init_logger(__name__)

# parse all the submitter arguments
parser = argparse.ArgumentParser()
parser.add_argument("--job_name", type=str, required=True)
parser.add_argument("--source_file", type=str, required=True)
parser.add_argument("--main_memory", type=float, required=True)
parser.add_argument("--main_time", type=float, required=True)
parser.add_argument("--main_scratch", type=float, required=True)
parser.add_argument("--function", type=str, required=True)
parser.add_argument("--executable", type=str, required=True)
parser.add_argument("--tasks", type=str, required=True)
parser.add_argument("--n_jobs", type=int, required=True)
parser.add_argument("--main_n_cores", type=int, required=True)
parser.add_argument("--log_dir", type=str, required=True)
parser.add_argument("--system", type=str, required=True)
parser.add_argument("--main_name", type=str, required=True)
parser.add_argument("--mode", type=str, required=True)
parser.add_argument("--main_mode", type=str, required=True)
parser.add_argument("--batchsize", type=int, required=True)
parser.add_argument("--max_njobs", type=int, required=True)
parser.add_argument("--discard_output", action="store_true", default=False)
parser.add_argument("--esub_verbosity", type=int, default=3)

args, function_args = parser.parse_known_args()
function = args.function
source_file = args.source_file
job_name = args.job_name
log_dir = args.log_dir
exe = args.executable
tasks = args.tasks
n_jobs = args.n_jobs
main_n_cores = args.main_n_cores
main_memory = args.main_memory
main_time = args.main_time
main_scratch = args.main_scratch
system = args.system
main_name = args.main_name
mode = args.mode
main_mode = args.main_mode

if mode == "mpi":
    try:
        from mpi4py import MPI

        is_master = MPI.COMM_WORLD.rank == 0
    except ImportError:
        if args.test:
            is_master = False
        else:
            raise ImportError(
                "You are attempting to run a MPI job. This requires a local "
                "MPI environement as well as the mpi4py package"
            )
else:
    is_master = True

logger_utils.set_logger_level(LOGGER, args.esub_verbosity)

# get path of log file and of file containing finished indices
path_log = utils.get_path_log(log_dir, job_name)
path_finished = utils.get_path_finished_indices(log_dir, job_name)

TIMEOUT_MESSAGE = (
    "Maximum number of pending jobs reached, "
    "will sleep for 30 minutes and retry"
)

# get rank of the processor
try:
    msg_limit_reached = "Pending job threshold reached."
    pipe_limit_reached = "stderr"
    if system == "bsub":
        rank = int(os.environ["LSB_JOBINDEX"])
    elif system == "slurm":
        rank = int(os.environ["SLURM_ARRAY_TASK_ID"])
    elif system == "daint":
        rank = 1
    rank -= 1
except KeyError:
    LOGGER.warning(
        "Unable to get current Job array index. Are you on a slurm or bsub "
        "system? Setting it to 0"
    )
    rank = 0

is_first = rank == 0

# Import the executable
executable = utils.import_executable(exe)

if is_master:
    if function == "main":
        LOGGER.info(
            f"Running the function {main_name} specified in executable"
        )
    else:
        LOGGER.info(
            "Running the function {} specified in executable".format(function)
        )

if (function == "rerun_missing") | (function == "check_missing"):
    if is_master:
        LOGGER.info("Checking if all main jobs terminated correctly...")

    indices_all = utils.get_indices(tasks)
    if function == "rerun_missing":
        indices_missing = utils.check_indices(
            indices_all,
            path_finished,
            executable,
            function_args,
            verbosity=args.esub_verbosity,
        )
    elif function == "check_missing":
        indices_missing = utils.check_indices(
            indices_all,
            path_finished,
            executable,
            function_args,
            check_indices_file=False,
            verbosity=args.esub_verbosity,
        )

    if is_master:
        utils.write_to_log(
            path_log, "Found {} missing indices".format(len(indices_missing))
        )

        if len(indices_missing) == 0:
            LOGGER.info("Nothing to resubmit. All jobs ended.")
        else:
            if len(indices_missing) > 1:
                # submit jobs
                tasks = ",".join(map(str, indices_missing[:-1]))
                n_jobs = len(indices_missing) - 1
                LOGGER.info(
                    "Re-Submitting tasks {} as {} jobs".format(tasks, n_jobs)
                )
                jobid = utils.submit_job(
                    tasks=tasks,
                    mode=main_mode,
                    exe=args.executable,
                    log_dir=log_dir,
                    function_args=function_args,
                    function="main_rerun",
                    source_file=source_file,
                    n_jobs=n_jobs,
                    job_name=job_name,
                    main_memory=main_memory,
                    main_time=main_time,
                    main_scratch=main_scratch,
                    main_n_cores=main_n_cores,
                    dependency="",
                    system=system,
                    main_name=main_name,
                    batchsize=args.batchsize,
                    discard_output=args.discard_output,
                    max_njobs=args.max_njobs,
                    verbosity=args.esub_verbosity,
                )

                utils.write_to_log(
                    path_log,
                    "Job id rerun_missing extended: {}".format(jobid[0]),
                )
            else:
                jobid = None

            # Change to local scratch if set; this has to be done
            # after submission,
            # s.t. that the pwd at submission time is
            # the original directory where the submission starts
            utils.cd_local_scratch(args.esub_verbosity)

            # run last job locally to not waste any resources
            index = indices_missing[-1]
            if is_master:
                LOGGER.info(
                    "##################### Starting Task {}\
                    #####################".format(
                        index
                    )
                )
            try:
                for index in getattr(executable, main_name)(
                    [index], function_args
                ):
                    if is_master:
                        utils.write_index(index, path_finished)
                if is_master:
                    LOGGER.info(
                        "##################### Finished Task {}\
                        #####################".format(
                            index
                        )
                    )
            except Exception as err:
                LOGGER.error("Error ocurred in rerun: {}".format(err))

            if is_master:
                if len(indices_missing) == 1:
                    utils.write_to_log(path_log, "First index is done")

            if is_master:
                # wait until all jobs are done
                if jobid is not None:
                    if system == "bsub":
                        while True:
                            while True:
                                output = dict(stdout=[], stderr=[])
                                with subprocess.Popen(
                                    shlex.split("bjobs {}".format(jobid[0])),
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    bufsize=1,
                                    universal_newlines=True,
                                ) as proc:

                                    # check for limit for maximum number
                                    # of pending jobs
                                    pending_limit_reached = False
                                    for line in getattr(
                                        proc, pipe_limit_reached
                                    ):

                                        pending_limit_reached = (
                                            msg_limit_reached in line
                                        )
                                        if pending_limit_reached:
                                            break
                                        else:
                                            output[pipe_limit_reached].append(
                                                line
                                            )

                                    # if the limit has been reached,
                                    # kill process and sleep
                                    if pending_limit_reached:
                                        proc.kill()
                                        LOGGER.warning(TIMEOUT_MESSAGE)
                                        time.sleep(60 * 30)
                                        continue

                                    # read rest of the output
                                    for line in proc.stdout:
                                        output["stdout"].append(line)
                                    for line in proc.stderr:
                                        output["stderr"].append(line)

                                    break

                            # check if process terminated successfully
                            if proc.returncode != 0:
                                raise RuntimeError(
                                    'Running the command "{}" failed with'
                                    "exit code {}. Error: \n{}".format(
                                        "bjobs {}".format(jobid[0]),
                                        proc.returncode,
                                        "\n".join(output["stderr"]),
                                    )
                                )

                            # check jobstate
                            jobstate = output["stdout"][-1].split()[2]
                            # for slurm either COMPLETED or FAILED?
                            if (jobstate == "DONE") | (jobstate == "EXIT"):
                                break
                            time.sleep(60)

                    else:
                        while True:
                            proc = subprocess.check_output(
                                shlex.split(f"sacct -j {jobid[0]} -o State")
                            ).decode()
                            proc = proc.replace(" ", "")
                            proc = proc.replace("State", "")
                            proc = proc.replace("-", "")
                            proc = proc.split("\n")
                            if ("RUNNING" not in proc) & (
                                "PENDING" not in proc
                            ):
                                break
                            time.sleep(60)

elif function == "merge_log_files":
    LOGGER.info("Merge all log files")
    indices_all = utils.get_indices(tasks)
    indices_missing = utils.check_indices(
        indices_all,
        path_finished,
        executable,
        function_args,
        verbosity=args.esub_verbosity,
    )
    completed = len(indices_all) - len(indices_missing)

    stdout_log_tot, stderr_log_tot = utils.get_log_filenames(
        log_dir, job_name, "main", system="tot"
    )
    # store previous log file without header
    # (e.g. when running rerun_missing independently from main)
    with open(stdout_log_tot, "r") as o:
        log_main = o.read().splitlines(True)[8:-10]
    with open(stderr_log_tot, "r") as e:
        log_err = e.read()

    # empty files
    open(stdout_log_tot, "w").close()
    open(stderr_log_tot, "w").close()

    message = (
        "#######################################################\n"
        "This is the merged main log file\n"
        f"{completed} / {len(indices_all)} jobs were successfully completed\n"
        f"Still missing are \n {indices_missing}\n"
        "#######################################################\n"
    )
    utils.save_write(stdout_log_tot, message)
    mainlog = (
        "#######################################################\n"
        "MAIN LOG FILES\n"
        "#######################################################\n"
    )
    utils.save_write(stdout_log_tot, mainlog)

    LOGGER.info("Start merging main files")
    utils.save_write(stdout_log_tot, "".join(log_main))
    utils.save_write(stderr_log_tot, "".join(log_err))
    for i in range(1, len(indices_all) + 1):
        stdout_log, stderr_log = utils.get_log_filenames(
            log_dir, job_name, "main", system
        )
        try:
            log_o = open(stdout_log % i, "r")
            log_e = open(stderr_log % i, "r")
            utils.save_write(stdout_log_tot, log_o.read())
            utils.save_write(stderr_log_tot, log_e.read())
            log_o.close()
            log_e.close()
            os.remove(stdout_log % i)
            os.remove(stderr_log % i)
        except Exception as err:
            LOGGER.error("Logfile for index {} not found".format(i))
            LOGGER.error("error message: {}".format(err))

    rer_log = (
        "#######################################################\n"
        "Rerun LOG FILES\n"
        "Note that the log of the last rerun index can be found \n"
        "in the rerun_missing log file.\n"
        "#######################################################\n"
    )
    utils.save_write(stdout_log_tot, rer_log)
    LOGGER.info("Start merging rerun files")
    LOGGER.info("This is also done when rerun_missing was not called")
    for i in range(1, len(indices_all) + 1):
        stdout_log, stderr_log = utils.get_log_filenames(
            log_dir, job_name, "main_r", system
        )
        try:
            log_o = open(stdout_log % i, "r")
            log_e = open(stderr_log % i, "r")
            utils.save_write(stdout_log_tot, log_o.read())
            utils.save_write(stderr_log_tot, log_e.read())
            log_o.close()
            log_e.close()
            os.remove(stdout_log % i)
            os.remove(stderr_log % i)
        except Exception as err:
            LOGGER.warning("Logfile for index {} not found".format(i))
            LOGGER.warning("error message: {}".format(err))
    utils.save_write(stdout_log_tot, message)
    LOGGER.info("Merged the log files")

else:
    # Change to local scratch if set
    utils.cd_local_scratch(args.esub_verbosity)

    # getting index list based on jobid
    indices = utils.get_indices_splitted(tasks, n_jobs, rank)

    if function == "main":
        is_first = rank == 0

        if is_master:
            LOGGER.info("Running on tasks: {}".format(indices))
        if system == "daint":
            try:
                for index in getattr(executable, main_name)(
                    indices, function_args
                ):
                    if is_master:
                        utils.write_index(index, path_finished)
                        LOGGER.info(
                            "##################### Finished Task {}\
                            #####################".format(
                                index
                            )
                        )

                    if is_first:
                        utils.write_to_log(path_log, "First index is done")
                        is_first = False
            except TypeError:
                # MPI does not terminate properly, therefore kill it
                sys.exit(0)
        else:
            for index in getattr(executable, main_name)(
                indices, function_args
            ):
                if is_master:
                    utils.write_index(index, path_finished)
                    LOGGER.info(
                        "##################### Finished Task {}\
                        #####################".format(
                            index
                        )
                    )

                if is_first:
                    utils.write_to_log(path_log, "First index is done")
                    is_first = False
    else:
        if is_master:
            utils.write_to_log(path_log, "Running {}".format(function))
            LOGGER.info(
                "Running {}, {} task(s), \
                        first: {}, last: {}".format(
                    function, len(indices), indices[0], indices[-1]
                )
            )
        getattr(executable, function)(indices, function_args)
        if is_master:
            utils.write_to_log(
                path_log, "Finished running {}".format(function)
            )



# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Joerg Herbel

import os
import shlex
import subprocess
import argparse
from ekit import logger as logger_utils

LOGGER = logger_utils.init_logger(__name__)


def get_job_ids(path_log):

    # Read file
    with open(path_log, "r") as f:
        lines = f.readlines()

    # Find job ids
    job_ids = []

    for line in lines:
        if "Job id" in line:

            job_id = ""
            line = line.strip()
            i = len(line) - 1

            while line[i].isdigit():
                job_id = line[i] + job_id
                i -= 1

            job_ids.append(job_id)

    return job_ids


def send_to_job(cmd, job_id):
    full_cmd = "{} {}".format(cmd, job_id)
    subprocess.call(shlex.split(full_cmd))


def main(dirpath_logs, cmd, log_filter=None):

    # Get all log files in directory
    filenames_logs = filter(
        lambda fn: not fn.startswith(".")
        and os.path.splitext(fn)[1] == ".log",
        os.listdir(dirpath_logs),
    )

    # Filter files
    if log_filter is not None:
        filenames_logs = filter(lambda fn: log_filter in fn, filenames_logs)

    # Extract job ids from files
    job_ids = []
    for filename in filenames_logs:
        job_ids += get_job_ids(os.path.join(dirpath_logs, filename))

    # Modify jobs
    for job_id in job_ids:
        LOGGER.info("Sending cmd {} to job {}".format(cmd, job_id))
        send_to_job(cmd, job_id)


if __name__ == "__main__":
    description = (
        "Send a command to all jobs logged in esub "
        "logfiles in a given directory"
    )
    parser = argparse.ArgumentParser(description=description, add_help=True)
    parser.add_argument(
        "--dirpath_logs",
        type=str,
        default="esub_logs",
        help="Directory containing esub logs",
    )
    parser.add_argument(
        "--cmd", type=str, required=True, help="Command to send to jobs"
    )
    parser.add_argument(
        "--log_filter",
        type=str,
        help="String that must be contained in log filename "
        "to receive cmd, default: all log files",
    )
    args = parser.parse_args()

    main(args.dirpath_logs, args.cmd, log_filter=args.log_filter)



# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Joerg Herbel

import os
import shutil
import shlex
import subprocess


def test_epipe():

    # create directory for test output
    path_testdir = os.path.join(os.getcwd(), 'esub_test_dir')
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    # run epipe
    cmd = 'epipe tests/pipe.yaml --ignore_jobid_errors --epipe_verbosity=4'
    subprocess.call(shlex.split(cmd))

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # check that log files were created and remove them
    log_dir = 'esub_logs'
    job_files = ['job1_done.dat',
                 'loopjob1__0_done.dat',
                 'loopjob1__1_done.dat',
                 'loopjob2__0_done.dat',
                 'loopjob2__1_done.dat']

    for job_file in job_files:
        assert os.path.isfile(os.path.join(log_dir, job_file))

    shutil.rmtree(log_dir)



# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Joerg Herbel

import os
import shutil
import shlex
import subprocess


def run_exec_example(
    test_dir, log_dir, mode="run", tasks_string=None, extra_esub_args=""
):

    # path to example file
    path_example = "example/exec_example.py"

    # build command
    cmd = (
        "esub {} --mode={} --output_directory={}"
        " --esub_verbosity=4 --log_dir={} {}".format(
            path_example, mode, test_dir, log_dir, extra_esub_args
        )
    )
    if tasks_string is not None:
        cmd += " --tasks={}".format(tasks_string)

    # main function
    subprocess.call(shlex.split(cmd))
    subprocess.call(shlex.split(cmd + " --function=main"))

    # rerun_missing
    subprocess.call(shlex.split(cmd + " --function=rerun_missing"))

    # watchdog
    subprocess.call(shlex.split(cmd + " --function=watchdog"))

    # merge
    subprocess.call(shlex.split(cmd + " --function=merge"))

    # all functions
    subprocess.call(shlex.split(cmd + " --function=all"))


def test_esub_run():

    # create directory for test output
    path_testdir = "esub_test_dir"
    cwd = os.getcwd()
    path_logdir = f"{cwd}/esub_test_dir_log"
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    # test with no tasks provided
    run_exec_example(path_testdir, path_logdir)

    # test with single task
    run_exec_example(path_testdir, path_logdir, tasks_string="99")

    # test with list of tasks
    run_exec_example(path_testdir, path_logdir, tasks_string="10,2,4")

    # test with range
    run_exec_example(path_testdir, path_logdir, tasks_string='"1 > 3"')

    # test run-tasks
    run_exec_example(
        path_testdir,
        path_logdir,
        mode="run-tasks",
        tasks_string='"0 > 4"',
        extra_esub_args="--n_jobs=2",
    )

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # test another example,
    # this time with renamed main function and missing merge function
    # works because main is renamed
    subprocess.call(
        shlex.split("esub example/exec_example.py --function=main")
    )
    # works because main is renamed
    subprocess.call(shlex.split("esub example/exec_example.py --function=all"))

    subprocess.call(
        shlex.split(
            "esub example/exec_example.py --function=main \
        --main_name=main_renamed"
        )
    )
    subprocess.call(shlex.split("rm randoms_0.npy"))
    subprocess.call(shlex.split("rm all_randoms.npy"))
    subprocess.call(shlex.split("rm -r esub_logs"))

    # check that log directory was created and remove it then
    assert os.path.isdir(path_logdir)
    shutil.rmtree(path_logdir)


def test_esub_jobarray():

    # create directory for test output
    path_testdir = "esub_test_dir_submit"
    cwd = os.getcwd()
    path_logdir = f"{cwd}/esub_test_dir_submit_log"
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    extra = (
        "--test --main_memory=50000 --main_time_per_index=100 "
        "--main_scratch=100000 --watchdog_memory=2400 --watchdog_time=50 "
        "--watchdog_scratch=90000 --merge_time=30 --merge_memory=98000 "
        "--merge_scratch=100000 --n_jobs=2 --main_n_cores_per_job=5 "
        "--merge_n_cores=10 --watchdog_n_cores=50 --max_njobs=100000"
    )
    # test with no tasks provided
    run_exec_example(
        path_testdir,
        path_logdir,
        mode="jobarray",
        extra_esub_args=extra,
        tasks_string='"1 > 3"',
    )

    # check if the submission strings are correct
    # check_cmd_strings('check_strings.txt', path_logdir)

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # check that log directory was created and remove it then
    assert os.path.isdir(path_logdir)
    shutil.rmtree(path_logdir)


def test_esub_mpi():

    # create directory for test output
    path_testdir = "esub_test_dir_mpi"
    cwd = os.getcwd()
    path_logdir = f"{cwd}/esub_test_dir_mpi_log"
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    extra = (
        "--test --main_memory=50000 --main_time=50 "
        "--main_scratch=100000 --watchdog_memory=2400 --watchdog_time=50 "
        "--watchdog_scratch=90000 --merge_time=30 --merge_memory=98000 "
        "--merge_scratch=100000 --n_jobs=2 --main_n_cores_per_job=100 "
        "--mpi_merge --mpi_watchdog"
    )
    # test with no tasks provided
    run_exec_example(
        path_testdir,
        path_logdir,
        mode="mpi",
        extra_esub_args=extra,
        tasks_string='"1 > 3"',
    )

    # check if the submission strings are correct
    # check_cmd_strings('check_strings_mpi.txt', path_logdir, mode='mpi')

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # check that log directory was created and remove it then
    assert os.path.isdir(path_logdir)
    shutil.rmtree(path_logdir)


def check_cmd_strings(file, path_logdir, mode="jobarray"):
    cwd = os.getcwd()

    sub_strings = []
    with open("{}/job.log".format(path_logdir), "r") as f:
        content = f.readlines()
    for line in content:
        if "bsub" in line:
            sub_strings.append("bsub" + line.split("bsub")[1])
    with open("tests/{}".format(file), "r") as f:
        check_strings_ = f.read().splitlines()
    check_strings = []
    for s in check_strings_:
        if len(s) > 0:
            check_strings.append(s.format(cwd, cwd, cwd, cwd))
    for ii, cmd_string in enumerate(sub_strings):
        if mode == "mpi":
            if (ii < 2) | (ii == 5):
                continue
        assert cmd_string == check_strings[ii]


def test_batching():

    # create directory for test output
    path_testdir = "esub_test_dir_submit_batch"
    cwd = os.getcwd()
    path_logdir = f"{cwd}/esub_test_dir_submit_batch_log"
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    extra = (
        "--test --main_memory=50000 --main_time_per_index=100 "
        "--main_scratch=100000 --watchdog_memory=2400 --watchdog_time=50 "
        "--watchdog_scratch=90000 --merge_time=30 --merge_memory=98000 "
        "--merge_scratch=100000 --n_jobs=11500 --batchsize=2000"
    )
    # test with no tasks provided
    run_exec_example(
        path_testdir,
        path_logdir,
        mode="jobarray",
        extra_esub_args=extra,
        tasks_string='"0 > 47875"',
    )

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # check that log directory was created and remove it then
    assert os.path.isdir(path_logdir)
    shutil.rmtree(path_logdir)



# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Dominik Zuercher

import os
import shutil
import subprocess
from tests.test_esub import run_exec_example


def test_esub_slurm_jobarray():

    # create directory for test output
    cwd = os.getcwd()
    path_testdir = 'esub_test_dir_submit_slurm'
    path_logdir = f'{cwd}/esub_test_dir_submit_slurm_log'
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    extra = '--test --main_memory=50000 --main_time_per_index=100 '\
            '--main_scratch=100000 --watchdog_memory=2400 --watchdog_time=50 '\
            '--watchdog_scratch=90000 --merge_time=30 --merge_memory=98000 '\
            '--merge_scratch=100000 --n_jobs=2 --main_n_cores_per_job=5 ' \
            '--system=slurm --merge_n_cores=10 --watchdog_n_cores=50 ' \
            '--max_njobs=100000 --keep_submit_files ' \
            '--additional_slurm_args=\"-C knl,--exclusive\"'
    # test with no tasks provided
    run_exec_example(path_testdir, path_logdir, mode='jobarray',
                     extra_esub_args=extra,
                     tasks_string='"1 > 3"')

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # check that log directory was created and remove it then
    assert os.path.isdir(path_logdir)
    shutil.rmtree(path_logdir)
    subprocess.call('rm submit*.slurm', shell=1)


def test_esub_mpi():

    # create directory for test output
    cwd = os.getcwd()
    path_testdir = 'esub_test_dir_mpi_slurm'
    path_logdir = f'{cwd}/esub_test_dir_mpi_slurm_log'
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    extra = '--test --main_memory=50000 --main_time_per_index=100 '\
            '--main_scratch=100000 --watchdog_memory=2400 --watchdog_time=50 '\
            '--watchdog_scratch=90000 --merge_time=30 --merge_memory=98000 '\
            '--merge_scratch=100000 --n_jobs=2 --main_n_cores_per_job=5 ' \
            '--system=slurm --merge_n_cores=10 --watchdog_n_cores=50 ' \
            '--max_njobs=100000 --keep_submit_files ' \
            '--additional_slurm_args=\"-C knl,--exclusive\"'

    # test with no tasks provided
    run_exec_example(path_testdir, path_logdir, mode='mpi',
                     extra_esub_args=extra,
                     tasks_string='"1 > 3"')

    # remove directory for test output
    shutil.rmtree(path_testdir)

    # check that log directory was created and remove it then
    assert os.path.isdir(path_logdir)
    shutil.rmtree(path_logdir)
    subprocess.call('rm submit*.slurm', shell=1)



# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Dominik Zuercher

import shutil
import shlex
import os
import subprocess


def test_check_logs():
    # setup mock esub_logs directory with log files
    path_testdir = 'esub_logs_check_logs'
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    # with unfinished jobs
    content = """2020-04-30 15:28:43.830214    Job id main: 120454427
2020-04-30 15:28:43.846407    Job id rerun_missing: 120454429
2020-04-30 15:28:43.849466    Job id merge: 120454431
    """

    with open('{}/job.log'.format(path_testdir), 'w+') as f:
        f.write(content)

    cmd = \
        'python esub/scripts/check_logs.py --dirpath_logs={}'.format(
            path_testdir)

    subprocess.call(shlex.split(cmd))

    # without unfinished jobs
    content = """2020-04-30 15:28:43.830214    Job id main: 120454427
2020-04-30 15:28:43.846407    Job id rerun_missing: 120454429
2020-04-30 15:28:43.849466    Job id merge: 120454431
2020-04-30 15:39:58.714086    First index is done
2020-04-30 15:54:44.332060    Found 0 missing indices
2020-04-30 15:54:44.358573    All indices finished
2020-04-30 15:55:32.758560    Running merge
2020-04-30 16:29:01.424964    Finished running merge
    """

    with open('{}/job.log'.format(path_testdir), 'w+') as f:
        f.write(content)

    cmd = \
        'python esub/scripts/check_logs.py --dirpath_logs={}'.format(
            path_testdir)

    subprocess.call(shlex.split(cmd))

    shutil.rmtree(path_testdir)


def test_send_cmd():
    # setup mock esub_logs directory with log files
    path_testdir = 'esub_logs_send_cmd'
    if not os.path.isdir(path_testdir):
        os.mkdir(path_testdir)

    # with unfinished jobs
    content = """2020-04-30 15:28:43.830214    Job id main: 120454427
2020-04-30 15:28:43.846407    Job id rerun_missing: 120454429
2020-04-30 15:28:43.849466    Job id merge: 120454431
    """

    with open('{}/job.log'.format(path_testdir), 'w+') as f:
        f.write(content)

    cmd = 'python esub/scripts/send_cmd.py --dirpath_logs={} ' \
        '--cmd=echo --log_filter=job'.format(
            path_testdir)

    # without unfinished jobs
    content = """2020-04-30 15:28:43.830214    Job id main: 120454427
2020-04-30 15:28:43.846407    Job id rerun_missing: 120454429
2020-04-30 15:28:43.849466    Job id merge: 120454431
2020-04-30 15:39:58.714086    First index is done
2020-04-30 15:54:44.332060    Found 0 missing indices
2020-04-30 15:54:44.358573    All indices finished
2020-04-30 15:55:32.758560    Running merge
2020-04-30 16:29:01.424964    Finished running merge
    """

    with open('{}/job.log'.format(path_testdir), 'w+') as f:
        f.write(content)

    cmd = 'python esub/scripts/send_cmd.py --dirpath_logs={} ' \
        '--cmd=echo --log_filter=job'.format(
            path_testdir)

    subprocess.call(shlex.split(cmd))
    shutil.rmtree(path_testdir)



#! /usr/bin/env python

# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Dominik Zuercher

# System imports
from __future__ import (
    print_function,
    division,
    absolute_import,
    unicode_literals,
)

# package import
import argparse
import math
import os
import sys
import collections
from esub import utils
from ekit import logger as logger_utils


LOGGER = logger_utils.init_logger(__name__)
TIMEOUT_MESSAGE = (
    "Maximum number of pending jobs reached, "
    "will sleep for 30 minutes and retry"
)


def starter_message():
    print(" ")
    print("                   ______   ")
    print("________________  ____  /_  ")
    print("_  _ \\_  ___/  / / /_  __ \\ ")
    print("/  __/(__  )/ /_/ /_  /_/ / ")
    print("\\___//____/ \\__,_/ /_.___/  ")
    print(" ")


def main(args=None):
    """
    Main function of esub.

    :param args: Command line arguments that are parsed
    """

    if args is None:
        args = sys.argv[1:]

    # initializing parser
    description = (
        "This is esub an user friendly and flexible tool to "
        "submit jobs to a cluster or run them locally"
    )
    parser = argparse.ArgumentParser(description=description, add_help=True)

    # default for resources
    resources = dict(
        main_memory=1000,
        main_time=4,
        main_time_per_index=0,
        main_scratch=2000,
        main_n_cores=1,
        watchdog_memory=1000,
        watchdog_time=4,
        watchdog_scratch=2000,
        watchdog_n_cores=1,
        merge_memory=1000,
        merge_time=4,
        merge_scratch=2000,
        merge_n_cores=1,
    )

    # parse all the submitter arguments
    parser.add_argument(
        "exec",
        type=str,
        help="path to the executable (python file "
        "containing functions main, watchdog, merge)",
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="run",
        choices=("run", "jobarray", "mpi", "run-mpi", "run-tasks"),
        help="The mode in which to operate. "
        "Choices: run, jobarray, mpi, run-mpi, "
        "run-tasks",
    )
    parser.add_argument(
        "--job_name",
        type=str,
        default="job",
        help="Individual name for this job. CAUTION: "
        "Multiple jobs with same name"
        "can confuse system!",
    )
    parser.add_argument(
        "--source_file",
        type=str,
        default="source_esub.sh",
        help="Optionally provide a source file which "
        "gets executed on the computing "
        "node before your jobs are started."
        "Useful for loading modules, "
        "declaring environemental variables and so on",
    )
    parser.add_argument(
        "--main_memory",
        type=float,
        default=resources["main_memory"],
        help="Memory allocated per core for main "
        "job in MB. Default: 1000 MB",
    )
    parser.add_argument(
        "--main_time",
        type=float,
        default=resources["main_time"],
        help="Job run time limit in hours for" " main job. Default 4h",
    )
    parser.add_argument(
        "--main_time_per_index",
        type=float,
        default=0,
        help="Job run time limit in hours for main "
        "job per index, overwrites main_time if set.",
    )
    parser.add_argument(
        "--main_scratch",
        type=float,
        default=resources["main_scratch"],
        help="Local scratch memory allocated "
        "for main job in MB. Default 2000 MB",
    )
    parser.add_argument(
        "--watchdog_memory",
        type=float,
        default=resources["watchdog_memory"],
        help="Memory allocated per core "
        "for watchdog job in MB. Default: 1000 MB",
    )
    parser.add_argument(
        "--watchdog_time",
        type=float,
        default=resources["watchdog_time"],
        help="Job run time limit in hours " "for watchdog job. Default: 4h",
    )
    parser.add_argument(
        "--watchdog_scratch",
        type=float,
        default=resources["watchdog_scratch"],
        help="Local scratch memory allocated "
        "for watchdog job. Default: 2000 MB",
    )
    parser.add_argument(
        "--merge_memory",
        type=float,
        default=resources["merge_memory"],
        help="Memory allocated per core for "
        "merge job in MB. Default: 1000 MB",
    )
    parser.add_argument(
        "--merge_time",
        type=float,
        default=resources["merge_time"],
        help="Job run time limit in hours " "for merge job. Default: 4h",
    )
    parser.add_argument(
        "--merge_scratch",
        type=float,
        default=resources["merge_scratch"],
        help="Local scratch memory allocated "
        "for merge job. Default: 2000 MB",
    )
    parser.add_argument(
        "--function",
        type=str,
        default="main",
        help="The functions that should be executed. "
        "Choices: main, watchdog, merge, "
        "rerun_missing, check_missing, all, "
        "or a list separated by whitespaces.",
    )
    parser.add_argument(
        "--main_name",
        type=str,
        default="main",
        help="Name of the main function in the executable.",
    )
    parser.add_argument(
        "--tasks",
        type=str,
        default="",
        help="Task string from which the indices are parsed. "
        "Either single index, list of "
        "indices (format 0,1,2 ), a range "
        "looking like int1 > int2 or path "
        "to a text file that holds "
        "a string in the format of the "
        "beforementioned formats.",
    )
    parser.add_argument(
        "--n_cores",
        type=int,
        default=-1,
        help="The number of cores to "
        "request for the main function (DEPRECATED).",
    )
    parser.add_argument(
        "--n_jobs",
        type=int,
        default=1,
        help="The number of jobs to " "request for the main function.",
    )
    parser.add_argument(
        "--max_njobs",
        type=int,
        default=-1,
        help="The maximal number of jobs that are allowed "
        "to run at the same time.",
    )
    parser.add_argument(
        "--dependency",
        type=str,
        default="",
        help="A dependency string that gets added to the "
        "dependencies (example after(<jobid>).",
    )
    parser.add_argument(
        "--system",
        type=str,
        default="bsub",
        choices=("bsub", "slurm", "daint"),
        help="Type of the system. "
        "Default is bsub (IBMs LSF system). "
        "To run single, large GPU jobs on Piz Daint "
        "use daint. "
        "Support for general slurm is expermimental!",
    )
    parser.add_argument(
        "--log_dir",
        type=str,
        default=os.path.join(os.getcwd(), "esub_logs"),
        help="Directory where to write the logs to. "
        "By default uses the current working directory.",
    )
    parser.add_argument(
        "--test",
        action="store_true",
        default=False,
        help="Test mode. Will not submit " "the jobs but only print.",
    )
    parser.add_argument(
        "--discard_output",
        action="store_true",
        default=False,
        help="If True all stdout/stderr is " "written to /dev/null.",
    )
    parser.add_argument(
        "--batchsize",
        type=int,
        default=100000,
        help="The maximum length of a jobarray. If requesting"
        " a larger jobarray it will automatically get "
        "split into smaller ones. Default: 100000",
    )
    parser.add_argument(
        "--additional_slurm_args",
        type=str,
        default="",
        help="Can pass comma-separated, additional, "
        "cluster-specific arguemnts. e.g. "
        '--additional_slurm_args="-C knl,--exclusive" '
        "to request "
        "only KNL nodes and to not share nodes with other "
        "users. The availablility of such arguments depends "
        "on the cluster you are using.",
    )
    parser.add_argument(
        "--additional_bsub_args",
        type=str,
        default="",
        help="Additional arguments for the LSF system. "
        "Example: '-R \"span[ptile=100]\"' to request that "
        "all 100 cores are on the same computing node",
    )
    parser.add_argument(
        "--esub_verbosity",
        type=int,
        default=3,
        choices=(0, 1, 2, 3, 4),
        help="esub verbosity. From 0-4 with 4 being "
        "the most verbose. Default is 3 (info).",
    )
    parser.add_argument(
        "--main_n_cores_per_job",
        type=int,
        default=resources["main_n_cores"],
        help="Number of cores per job for the main function (DEPRECATED). "
        "Default: 1",
    )
    parser.add_argument(
        "--main_n_cores",
        type=int,
        default=resources["main_n_cores"],
        help="Number of cores per job for the main function. " "Default: 1",
    )
    parser.add_argument(
        "--watchdog_n_cores",
        type=int,
        default=resources["watchdog_n_cores"],
        help="Number of cores per job for the watchdog. "
        "fuction. Default: 1",
    )
    parser.add_argument(
        "--merge_n_cores",
        type=int,
        default=resources["merge_n_cores"],
        help="Number of cores per job for the merge. " "Default: 1",
    )
    parser.add_argument(
        "--mpi_merge",
        action="store_true",
        default=False,
        help="If True merge function is run as an MPI job. "
        "Otherwise as a normal job with merge_cores.",
    )
    parser.add_argument(
        "--mpi_watchdog",
        action="store_true",
        default=False,
        help="If True watchdog function is run as an MPI job. "
        "Otherwise as a normal job with watchdog_cores.",
    )
    parser.add_argument(
        "--keep_submit_files",
        action="store_true",
        default=False,
        help="If True does not delete the SLURM submission "
        "Ignored in bsub mode.",
    )
    parser.add_argument(
        "--nodes",
        type=int,
        default=1,
        help="Number of GPU nodes to allocate. " "Only for daint. Default: 1",
    )
    parser.add_argument(
        "--MPI_tasks_per_core",
        type=int,
        default=1,
        help="Number of MPI tasks per host CPU core. "
        "Values larger than 1 lead to hyperthreading. "
        "Only for daint. Default: 1",
    )
    parser.add_argument(
        "--MPI_tasks_per_node",
        type=int,
        default=1,
        help="Number of MPI tasks per GPU node. " "Only for daint. Default: 1",
    )
    parser.add_argument(
        "--OpenMP_threads_per_task",
        type=int,
        default=1,
        help="Number of OpenMP threads per task. "
        "Only for daint. Default: 1",
    )

    args, function_args = parser.parse_known_args(args)

    # Additional parser to detect which arguments are passed via commandline,
    # This is necessary if the commandline argument is different from the one given
    # in the resource function but the same as the esub default
    # https://stackoverflow.com/questions/32056910/
    aux_parser = argparse.ArgumentParser(argument_default=argparse.SUPPRESS)
    for arg in vars(args):
        if type(getattr(args, arg)) != bool:
            # store_true would cause problems
            aux_parser.add_argument("--" + arg)
    commandline_args, _ = aux_parser.parse_known_args()

    logger_utils.set_logger_level(LOGGER, args.esub_verbosity)

    if args.n_cores > -1:
        LOGGER.warning(
            "DEPRECATION WARNING: The n_cores option will be dropped in a "
            "future version. Use n_jobs instead."
        )
        args.n_jobs = args.n_cores

    if args.main_n_cores_per_job > 1:
        LOGGER.warning(
            "DEPRECATION WARNING: The main_n_cores_per_job option will be "
            "dropped in a future version. Use main_n_cores instead."
        )
        args.main_n_cores = args.main_n_cores_per_job

    # Make log directory if non existing
    log_dir = args.log_dir
    if not os.path.isdir(log_dir):
        os.makedirs(log_dir)
        LOGGER.debug("Created log directory {}".format(log_dir))

    mode = args.mode
    add_args = args.additional_slurm_args
    main_name = args.main_name
    job_name = args.job_name
    source_file = args.source_file
    tasks = args.tasks
    exe = args.exec
    n_jobs = args.n_jobs
    main_n_cores = args.main_n_cores
    keep_submit_files = args.keep_submit_files
    function = args.function
    system = args.system
    nodes = args.nodes
    MPI_tasks_per_core = args.MPI_tasks_per_core
    MPI_tasks_per_node = args.MPI_tasks_per_node
    OpenMP_threads_per_task = args.OpenMP_threads_per_task

    if system == "daint":
        LOGGER.warning(
            "\n============================================================ \n"
            "You are requesting a GPU job on the Piz Daint cluster. \n"
            "Piz Daint allows to run hybrid MPI+OpenMP jobs. \n"
            "Each node on Piz Daint features a 12 core XC50 Intel host CPU "
            "and a NVIDIA Tesla P100 GPU (16 GB VRAM). \n"
            "The normal esub workflow is not appropriate for Piz Daint! \n"
            "Only single task main jobs are allowed and they will always "
            "run in MPI mode. \n"
            "The n_jobs, tasks and main_n_cores_per_job arguments "
            "are ignored. \n"
            "Instead use the nodes, MPI_tasks_per_core, MPI_tasks_per_node, "
            "OpenMP_threads_per_task arguments. \n"
            "=============================================================="
        )
        if function != "main":
            raise ValueError(
                "Functions different from main are not "
                "supported on Piz Daint."
            )
        if mode != "mpi":
            raise ValueError(
                "Modes different from mpi are not " "supported on Piz Daint."
            )
    ext_dependencies = args.dependency

    # Make sure that executable exits
    if os.path.isfile(exe):
        if not os.path.isabs(exe):
            exe = os.path.join(os.getcwd(), exe)
            LOGGER.debug(f"Using executable file at {exe}")
    else:
        raise FileNotFoundError(
            "Did not find {}. Please specify a valid path for "
            "executable".format(exe)
        )

    starter_message()

    # Set path to log file and to file storing finished main job ids
    path_log = utils.get_path_log(log_dir, job_name)
    LOGGER.debug(f"Using log file at {path_log}")

    path_finished = utils.get_path_finished_indices(log_dir, job_name)
    LOGGER.debug(f"Storing finished indices at {path_finished}")
    LOGGER.info("Running in run mode {}".format(mode))

    # importing the functions from the executable
    executable = utils.import_executable(exe)

    # check if required function exists. Otherwise skip it
    if "," in function:
        function = function.split(",")
    else:
        function = [function]
    if (len(function) == 1) and (function[0] == "all"):
        function = "all"
    if function == "all":
        function = ["main", "rerun_missing", "watchdog", "merge"]
    for func in function:
        if (func == "rerun_missing") | (func == "merge_log_files"):
            continue
        elif func == "main":
            if not hasattr(executable, main_name):
                LOGGER.warning(
                    "Did not find main function {} in the executable. "
                    "Skipping it...".format(main_name)
                )
                function.remove(func)
        else:
            if not hasattr(executable, func):
                LOGGER.warning(
                    "Did not find function {} in the executable. "
                    "Skipping it...".format(func)
                )
                function.remove(func)
    if len(function) == 0:
        LOGGER.warning("No function to run found. Exiting.")
        sys.exit(0)

    # run setup if implemented
    if hasattr(executable, "setup"):
        LOGGER.info("Running setup function from executable")
        getattr(executable, "setup")(function_args)

    # run get_tasks function if implemented
    if system == "daint":
        LOGGER.warning("Setting number of jobs to 1!")
        tasks = "0"
    else:
        if len(tasks) == 0:
            if hasattr(executable, "get_tasks"):
                LOGGER.info("Running get_tasks function from executable")
                tasks_ = getattr(executable, "get_tasks")(function_args)
                # convert list or range to string
                if isinstance(tasks_, list):
                    tasks = ""
                    for t in tasks_:
                        tasks += f"{t},"
                    tasks = tasks[:-1]
                elif isinstance(tasks_, tuple) & (len(tuple) == 2):
                    tasks = f"{int(tasks_[0])} > {int(tasks_[1])}"
                else:
                    raise ValueError(
                        "Your get_tasks function returned a value that is "
                        "not allowed. "
                        "Needs to return a list of integers or a tuple with "
                        "two entries "
                        "indicating the first and last (exclusive) "
                        "index to run."
                    )
            else:
                # default
                tasks = "0"

    # get resources from executable if implemented
    res_update = dict()
    if hasattr(executable, "resources"):
        LOGGER.info(
            "Running resources function from executable. "
            "Updating resource requirements."
        )
        res_update = getattr(executable, "resources")(function_args)

    # overwrite with non-default command-line input
    for res_name in res_update.keys():
        if hasattr(commandline_args, res_name):
            commandline_val = getattr(commandline_args, res_name)
            LOGGER.debug(
                f"Overriding resource {res_name} from resource function: "
                f"{res_update[res_name]} -> {commandline_val}"
            )
            res_update[res_name] = commandline_val
    resources.update(res_update)

    if resources["main_time_per_index"] > 0:
        n_indices = len(utils.get_indices(tasks))
        resources["main_time"] = resources["main_time_per_index"] * math.ceil(
            n_indices / n_jobs
        )
        LOGGER.debug(
            f"main_time_per_index is set -> Overriding "
            f"main_time to {resources['main_time']}h"
        )
    del resources["main_time_per_index"]

    # check if log files should be overwritten
    overwrite_log = (function == "all") | (function == "main")

    # CASE 1 : run locally
    if (mode == "run") | (mode == "run-mpi") | (mode == "run-tasks"):
        LOGGER.info("Running locally!")

        # adding function and tasks arguments
        if function == "all":
            LOGGER.debug("Running all functions sspecified in executable")
        else:
            LOGGER.debug(
                "Running the function(s) {} "
                "specified in executable".format(", ".join(function))
            )

        # getting index list
        indices = utils.get_indices(tasks)
        LOGGER.info("Running on tasks: {}".format(indices))

        # loop over functions
        for f in function:

            LOGGER.info(f"Running function {f}")
            indices_use = indices

            # check if function is specified
            if f == "main" or f == "rerun_missing":
                function_found = hasattr(executable, main_name)
            elif f == "merge_log_files":
                function_found = True
            else:
                function_found = hasattr(executable, f)

            if not function_found:
                LOGGER.warning(
                    "The requested function {} is missing in the executable. "
                    "Skipping...".format(f)
                )
                continue

            if f == "main":
                # resetting missing file
                LOGGER.debug("Resetting file holding finished indices")
                utils.robust_remove(path_finished)

            if f == "rerun_missing":
                indices_use = utils.check_indices(
                    indices,
                    path_finished,
                    executable,
                    function_args,
                    verbosity=args.esub_verbosity,
                )
                if len(indices_use) > 0:
                    LOGGER.info("Rerunning tasks: {}".format(indices_use))
                    f = "main"
                else:
                    LOGGER.info("All indices are finished, nothing to re-run.")
                    continue

            if f == "check_missing":
                indices_use = utils.check_indices(
                    indices,
                    path_finished,
                    executable,
                    function_args,
                    check_indices_file=False,
                    verbosity=args.esub_verbosity,
                )
                if len(indices_use) > 0:
                    LOGGER.info("Rerunning tasks: {}".format(indices_use))
                    f = "main"
                else:
                    LOGGER.info("All indices are finished, nothing to re-run.")
                    continue

            if f == "main":
                if mode == "run":
                    for index in getattr(executable, main_name)(
                        indices, function_args
                    ):
                        LOGGER.info(
                            "##################### Starting Task {} "
                            "#####################".format(index)
                        )
                        utils.write_index(index, path_finished)
                        LOGGER.info(
                            "##################### Finished Task {} "
                            "#####################".format(index)
                        )

                elif mode == "run-mpi":
                    LOGGER.info(
                        "##################### Starting MPI job"
                        "#####################"
                    )
                    utils.run_local_mpi_job(
                        exe,
                        main_n_cores,
                        function_args,
                        LOGGER,
                        main_name,
                        args.esub_verbosity,
                    )
                    LOGGER.info(
                        "##################### Finished MPI job"
                        "#####################"
                    )

                elif mode == "run-tasks":
                    LOGGER.info(
                        "##################### Starting parallel tasks {} "
                        "#####################".format(tasks)
                    )
                    dones = utils.run_local_tasks(
                        executable, n_jobs, function_args, tasks, main_name
                    )
                    for index in dones:
                        utils.write_index(index, path_finished)
                        LOGGER.info(
                            "##################### Finished Task {} "
                            "#####################".format(index)
                        )

            else:
                getattr(executable, f)(indices_use, function_args)

    # CASE 2 and 3 : running jobs on cluster (MPI or jobarray)
    elif (mode == "jobarray") | (mode == "mpi") | (mode == "tasks"):
        # Add dependencies to functions
        if (function == "all") & (mode == "jobarray"):
            function = ["main", "watchdog", "rerun_missing", "merge"]
            LOGGER.info(
                "Submitting all functions specified in executable "
                "to queuing system. Watchdog running along "
                "main. Trying to rerun jobs after main finished. "
                "Merge running at the end."
            )
        elif (function == "all") & ((mode == "mpi") | (mode == "tasks")):
            function = ["main", "watchdog", "merge"]
            LOGGER.info(
                "Submitting all functions specified in executable "
                "to queuing system. Watchdog running along "
                "main. Merge running at the end."
            )
        else:
            LOGGER.info(
                "Submitting the function(s) {} specified in "
                "executable to queuing system".format(", ".join(function))
            )

        if (system == "slurm") & (mode == "jobarray"):
            # add merge_log_files for the main function
            if ("main" in function) | ("rerun_missing" in function):
                function.append("merge_log_files")

        jobids = collections.OrderedDict()
        for ii, f in enumerate(function):
            if (
                (f == "main")
                or (f == "rerun_missing")
                or (f == "check_missing")
            ):
                function_found = hasattr(executable, main_name)
            elif f == "merge_log_files":
                function_found = True
            else:
                function_found = hasattr(executable, f)

            if not function_found:
                LOGGER.warning(
                    "The requested function {} is missing in the executable. "
                    "Skipping...".format(f)
                )
                continue

            if f == "main":
                # resetting missing file
                LOGGER.debug("Resetting file holding finished indices")
                utils.robust_remove(path_finished)
                n_jobs_use = n_jobs
                mode_ = mode
            elif f == "watchdog":
                n_jobs_use = 1
                if args.mpi_watchdog:
                    mode_ = "mpi"
                else:
                    mode_ = "jobarray"
            elif f == "merge":
                n_jobs_use = 1
                if args.mpi_merge:
                    mode_ = "mpi"
                else:
                    mode_ = "jobarray"
            else:
                # reruns
                n_jobs_use = 1
                mode_ = mode

            LOGGER.debug(
                f"Submitting function {f} broken down "
                f"into {n_jobs_use} job(s)"
            )

            # reset logs
            LOGGER.debug("Resetting log files")
            stdout_log, stderr_log = utils.get_log_filenames(
                log_dir, job_name, f
            )
            utils.robust_remove(stdout_log)
            utils.robust_remove(stderr_log)

            # the current job depends at most on the previous one
            # (e.g., rerun_missing does not need to wait for the
            # watchdog to finish)
            dependency = utils.get_dependency_string(
                f,
                jobids,
                ext_dependencies,
                system,
                verbosity=args.esub_verbosity,
            )

            jobid = utils.submit_job(
                tasks,
                mode_,
                exe,
                log_dir,
                function_args,
                function=f,
                source_file=source_file,
                n_jobs=n_jobs_use,
                job_name=job_name,
                dependency=dependency,
                system=system,
                main_name=main_name,
                test=args.test,
                add_args=add_args,
                batchsize=args.batchsize,
                max_njobs=args.max_njobs,
                add_bsub=args.additional_bsub_args,
                discard_output=args.discard_output,
                verbosity=args.esub_verbosity,
                main_mode=mode,
                keep_submit_files=keep_submit_files,
                nodes=nodes,
                MPI_tasks_per_core=MPI_tasks_per_core,
                MPI_tasks_per_node=MPI_tasks_per_node,
                OpenMP_threads_per_task=OpenMP_threads_per_task,
                **resources,
            )
            jobids[f] = jobid

            LOGGER.info(
                "Submitted job for function {}. "
                "Got jobid(s) {}".format(f, jobid)
            )

        jobid_str = ""
        for jobid_list in jobids.values():
            if isinstance(jobid_list, int):
                jobid_str += f"{str(jobid_list)} "
            elif isinstance(jobid_list, list):
                for id in jobid_list:
                    jobid_str += f"{str(id)} "
            else:
                pass
        if len(jobid_str) == 0:
            jobid_str = "None"

        LOGGER.info("Submission finished")
        print(f"Submitted jobids: {jobid_str}")

        # write to log
        if overwrite_log:
            utils.write_to_log(
                path_log, "esub arguments: \n{}".format(args), mode="w"
            )
            utils.write_to_log(
                path_log, "function arguments: \n{}".format(function_args)
            )

        for fun, jobid in jobids.items():
            utils.write_to_log(path_log, "Job id {}: {}".format(fun, jobid))


if __name__ == "__main__":
    main()



#! /usr/bin/env python

# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Dominik Zuercher

# System imports
from __future__ import (print_function, division, absolute_import,
                        unicode_literals)

# package imports
import sys
import argparse
import subprocess
import shlex
import yaml
from ekit import logger as logger_utils

from esub import esub


LOGGER = logger_utils.init_logger(__name__)


def starter_message():
    print(' ')
    print("             _____               ")
    print("________________(_)____________  ")
    print("_  _ \\__  __ \\_  /___  __ \\  _ \\ ")
    print("/  __/_  /_/ /  / __  /_/ /  __/ ")
    print("\\___/_  .___//_/  _  .___/\\___/  ")
    print("     /_/          /_/            ")
    print(' ')


def format_loop_cmd(command, index, entry):
    """
    Inserts a loop index into a command if specified by eg. "{}"
    inside the command. Then evaluates the command
    as an f-string to resolve possible (e.g. mathematical) expressions.

    :param command: command string
    :param index: loop index
    :return: command with loop index inserted at all places specified by
             curly brackets and optional format
             specifications and afterwards as f-string evaluated
    """

    # format
    index_repeated = []

    while True:
        try:
            # if this fails, there are more places where the index
            # should be inserted than we currently have items in
            # the list
            command_formatted = command.format(*index_repeated)
            break
        except IndexError:
            index_repeated.append(entry)

    # evaluate as f-string
    command_formatted = eval('f"{}"'.format(command_formatted))

    return command_formatted


def represents_int(s):
    """
    Checks if string can be converted to integer

    :param s: string to be checked
    :return: True if s can be converted to an integer and False otherwise
    """
    try:
        int(s)
        return True
    except ValueError:
        return False


def make_submit_command(base_cmd, name, deps, job_dict, parameters):
    """
    Adds the dependencies to the base command given in the pipeline
    Also tries to replace variables indicated by [] brackets.

    :param base_cmd: The command string given in pipeline file
    :param name: The job name
    :param deps: The dependencies of the job
    :param job_dict: The jobs dictionary
    :param parameters: A dictionary holding global variables.
                       Function attempts to replace them in the command string
    :return: Complete submission string
    """

    dep_string = ''
    for dep in deps:
        if represents_int(dep):
            job_ids = [dep]
        elif dep in job_dict.keys():
            job_ids = job_dict[dep]
        else:
            LOGGER.warning(
                "Did not find a job named {}. \
                 Ignoring dependencies on it".format(dep))
            continue
        for job_id in job_ids:
            dep_string += 'ended({}) && '.format(int(job_id))
    dep_string = dep_string[:-4]
    dep_string = '"' + dep_string + '"'

    cmd = base_cmd + \
        ' --job_name={}'.format(name) + ' --dependency={}'.format(dep_string)

    # Attempt variable replacement
    while True:
        rep = cmd.find('$')
        if rep == -1:
            break
        start = cmd[rep:].index('[') + rep
        stop = cmd[rep:].index(']') + rep
        rep_key = cmd[start + 1: stop]
        cmd = cmd[:rep] + str(parameters[rep_key]) + cmd[stop + 1:]
    return cmd


def submit(cmd, assert_ids=True):
    """
    Runs the command and receives its job index
    which gets added to the dependency list.

    :param cmd: Command string to submit
    :param assert_ids: whether to check that job ids were successfully obtained
    :return: The ids of the just submitted jobs
    """

    stdout_lines = []
    with subprocess.Popen(shlex.split(cmd),
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          bufsize=1,
                          universal_newlines=True) as proc:

        # read stdout
        for line in proc.stdout:
            stdout_lines.append(line)

            if esub.TIMEOUT_MESSAGE in line:
                LOGGER.warning(esub.TIMEOUT_MESSAGE)
        # read stderr
        for line in proc.stdout:
            stdout_lines.append(line)

            if esub.TIMEOUT_MESSAGE in line:
                LOGGER.warning(esub.TIMEOUT_MESSAGE)

        # read stderr
        stderr = '\n'.join(proc.stderr)

    # raise an exception in case the process did not terminate successfully
    if proc.returncode != 0:
        raise RuntimeError('Running the command \"{}\" failed with\
                            exit code {}. Error: \n{}'. format(
            cmd, proc.returncode, stderr))

    # get ids of newly submitted jobs
    last_line = stdout_lines[-1]

    ids = []
    for string in last_line.split('Submitted job ids:')[-1].split(' '):
        try:
            int(string)
            ids.append(string)
        except ValueError:
            pass

    if assert_ids:
        assert len(ids) > 0, 'Something went wrong, did not manage ' \
                             'to get any job ids from stdout'
    return ids


def get_parameters(step):
    """
    Extracts and stores global variables from the\
    parameter object in epipe file

    :param step: epipe object instance
    :return : Dictionary holding the global variables
    """
    param_list = step['parameters']
    parameters = {}
    for element in param_list:
        key = list(element.keys())[0]
        parameters[key] = element[key]

    parameter_dict = {}
    for par_key in parameters.keys():
        parameter_dict[par_key] = parameters[par_key]

    for key in parameter_dict.keys():
        LOGGER.debug(f"Setting parameter {key} -> {parameter_dict[key]}")
    return parameter_dict


def process_item(step, job_dict, index=-1, loop_entry=None,
                 loop_dependence=None,
                 parameters=None, assert_ids=True):
    """
    Processes one of the epipe items in the pipeline

    :param step: The job item
    :param job_dict: The previous job id dictionary
    :param index: Loop index if within a loop
    :param loop_dependence: The overall dependencies of the
                            loop structure if within a loop
    :param parameters: A dictionary holding global variables. Function
                       attempts to replace them in the command string
    :param assert_ids: whether to check that job ids were successfully obtained
    :return: An updated dictionary containing the already
             submitted jobs and the newly submitted job
    """

    if parameters is None:
        parameters = dict()

    deps = []
    in_loop = index >= 0

    # direct dependencies of the item itself
    if 'dep' in step.keys():
        step['dep'] = str(step['dep'])
        dependencies = step['dep'].replace(' ', '').split(',')
        for dep in dependencies:
            deps.append(dep)

        # if we are inside a loop, then format the
        # loop-internal dependencies as necessary
        if in_loop:
            for ii, dep in enumerate(deps):
                if not represents_int(dep):
                    deps[ii] = '{}__{}'.format(dep, loop_entry)

    # add the global dependencies of the loop
    if loop_dependence is not None:
        for lp in loop_dependence:
            deps.append(lp)

    # get the job name and the job command
    job_name = step['name']
    base_cmd = step['cmd']
    if in_loop:
        job_name = '{}__{}'.format(job_name, loop_entry)
        base_cmd = format_loop_cmd(base_cmd, index, loop_entry)

    # submit the job
    LOGGER.debug(f"Propagating dependencies {deps} to job {job_name}")
    LOGGER.info('Submitting job {}'.format(job_name))
    cmd = make_submit_command(base_cmd, job_name, deps, job_dict, parameters)
    print(cmd)
    new_ids = submit(cmd, assert_ids=assert_ids)
    job_dict[job_name] = new_ids


def main(args=None):
    """
    epipe main function. Accepts a epipe yaml
    configuration file and runs the pipeline.

    :param args: Command line arguments
    """

    if args is None:
        args = sys.argv[1:]

    description = "This is epipe a tool to easily submit\
                   pipelines to a clusters queing system"
    parser = argparse.ArgumentParser(description=description, add_help=True)
    # parse all the submitter arguments
    parser.add_argument('pipeline', type=str, action='store',
                        help='Path to the pipeline file. Format should be as '
                        'specified in the epipe example. Check '
                        'documentations')
    parser.add_argument('--ignore_jobid_errors', action='store_true',
                        help='Switches off checks checking '
                        'wheter ids of submitted '
                        'jobs were successfully obtained, useful for '
                        'testing')
    parser.add_argument('--epipe_verbosity', type=int,
                        default=3, choices=(0, 1, 2, 3, 4),
                        help='epipe verbosity. From 0-4 with 4 being '
                        'the most verbose. Default is 3 (info).')

    args = parser.parse_args(args)
    logger_utils.set_logger_level(LOGGER, args.epipe_verbosity)

    pipeline = args.pipeline
    LOGGER.debug(f"Using pipeline in file {pipeline}")

    assert_ids = not args.ignore_jobid_errors

    # read pipeline file
    with open(pipeline, 'r') as f:
        pipeline = yaml.load(f, yaml.FullLoader)

    starter_message()

    # parse all the jobs and their dependencies from the pipeline string
    job_dict = {}

    LOGGER.info('Starting submission')

    parameters = None

    for step in pipeline:
        LOGGER.debug(f"Processing item {step}")
        if 'parameters' in step.keys():
            LOGGER.debug("Found parameter object")
            parameters = get_parameters(step)
            continue

        if 'loop' in step.keys():
            LOGGER.debug("Found loop object")

            loop_entries = list(step['loop'])
            if len(loop_entries) == 2:
                if isinstance(loop_entries[0], int) & isinstance(
                   loop_entries[1], int):
                    LOGGER.warning(
                        "Loop object has structure "
                        "[int1, int2] -> Intrepreting as a range!")
                    loop_entries = [x for x in range(
                        loop_entries[0], loop_entries[1])]
            LOGGER.info(f"Submitting loop indices {loop_entries}")
            deps = []
            # get the names of all jobs that have been
            # submitted before the loop
            submitted_jobs = set(job_dict.keys())

            if 'dep' in step.keys():
                step['dep'] = str(step['dep'])
                dependencies = step['dep'].replace(' ', '').split(',')
                for dep in dependencies:
                    deps.append(dep)

            for index, entry in enumerate(loop_entries):
                for item in step['items']:
                    process_item(item, job_dict, index=index, loop_entry=entry,
                                 loop_dependence=deps, parameters=parameters,
                                 assert_ids=assert_ids)

            # find the names of the newly submitted jobs
            loop_jobs = set(job_dict.keys()) - submitted_jobs

            # add the loop as a whole to the job dictionary
            job_dict[step['name']] = []
            for loop_job in loop_jobs:
                job_dict[step['name']] += job_dict[loop_job]

        else:
            process_item(step, job_dict, parameters=parameters,
                         assert_ids=assert_ids)

    LOGGER.info('Submission finished')


if __name__ == '__main__':
    main()



# Copyright (C) 2019 ETH Zurich,
# Institute for Particle Physics and Astrophysics
# Author: Joerg Herbel

import os
import argparse

from ekit import logger as logger_utils

LOGGER = logger_utils.init_logger(__name__)


def main(dirpath_logs):

    # Get all log files in directory
    filenames_logs = filter(
        lambda fn: not fn.startswith(".")
        and os.path.splitext(fn)[1] == ".log",
        os.listdir(dirpath_logs),
    )
    filenames_logs = sorted(filenames_logs)

    # Check
    unfinished_logs = []

    for filename in filenames_logs:

        # read log file
        with open(os.path.join(dirpath_logs, filename), mode="r") as f:
            log = f.read()

        # check which functions were submitted
        main_run = "Job id main" in log
        rerun_missing = "Job id rerun_missing" in log
        merge_run = "Job id merge" in log
        watchdog_run = "Job id watchdog" in log

        # check if main job only and MPI mode
        if main_run and not (rerun_missing | merge_run | watchdog_run):
            if "Finished running main" in log:
                continue

        # check jobarray case
        expected_content = []
        if main_run or rerun_missing:
            expected_content.append("All indices finished")
        if merge_run:
            expected_content.append("Finished running merge")
        if watchdog_run:
            expected_content.append("Finished running watchdog")

        for exp_cont in expected_content:
            if exp_cont not in log:
                unfinished_logs.append(filename)
                break

    # Report
    if len(unfinished_logs) == 0:
        LOGGER.info("No unfinished jobs found")
    else:
        LOGGER.info(
            "Logs containing unfinished jobs:\n{}".format(
                "\n".join(unfinished_logs)
            )
        )


if __name__ == "__main__":
    description = (
        "Send a command to all jobs logged in "
        "esub logfiles in a given directory"
    )
    parser = argparse.ArgumentParser(description=description, add_help=True)
    parser.add_argument(
        "--dirpath_logs",
        type=str,
        default="esub_logs",
        help="Directory containing esub logs",
    )
    args = parser.parse_args()
    main(args.dirpath_logs)
